/*
 * Autogenerated by the Meson build system.
 * Do not edit, your changes will be lost.
 */

#pragma once

#define ARCH_AARCH64 1

#define ARCH_ARM 0

#define ARCH_PPC64LE 0

#define ARCH_X86 0

#define ARCH_X86_32 0

#define ARCH_X86_64 0

#define CONFIG_16BPC 1

#define CONFIG_8BPC 1

#define CONFIG_LOG 1

#define ENDIANNESS_BIG 0

#define HAVE_ASM 0

#define HAVE_AS_FUNC 0

#define HAVE_CLOCK_GETTIME 1

#define HAVE_POSIX_MEMALIGN 1

#define HAVE_UNISTD_H 1

#define PIC 3

#define PREFIX 1

